import subprocess
import tempfile
from pathlib import Path
from typing import Sequence, List, Dict

import mrcfile
import numpy as np

from yet_another_imod_wrapper.utils.io import write_adoc
from yet_another_imod_wrapper.etomo_directory import EtomoDirectory


def prepare_etomo_directory(
        directory: Path,
        tilt_series: np.ndarray,
        tilt_angles: Sequence[float],
        basename: str,
) -> EtomoDirectory:
    """Prepare a directory for IMOD tilt-series alignment."""
    directory.mkdir(exist_ok=True, parents=True)
    directory = EtomoDirectory(basename=basename, directory=directory)
    mrcfile.write(str(directory.tilt_series_file), tilt_series.astype(np.float32))
    np.savetxt(directory.rawtlt_file, tilt_angles, fmt='%.2f', delimiter='')
    return directory


def run_batchruntomo(
        directory: Path, basename: str, directive: Dict[str, str]
) -> None:
    """Run batchruntomo on a single tilt-series with a specified directive."""
    with tempfile.TemporaryDirectory() as temporary_directory:
        directive_file = Path(temporary_directory) / 'directive.adoc'
        write_adoc(directive, directive_file)
        batchruntomo_command = _get_batchruntomo_command(
            directory=directory,
            basename=basename,
            directive_file=directive_file
        )
        subprocess.run(batchruntomo_command)


def _get_batchruntomo_command(
        directory: Path, basename: str, directive_file: Path
) -> List[str]:
    """Get batchruntomo command."""
    command = [
        'batchruntomo',
        '-DirectiveFile', f'{directive_file}',
        '-CurrentLocation', f'{directory}',
        '-RootName', basename,
        '-EndingStep', '6'
    ]
    return command



from os import PathLike
from pathlib import Path
from typing import Dict, Any, List

import numpy as np

from .etomo_directory import EtomoDirectory
from .utils.io import read_adoc
from .constants import TARGET_PIXEL_SIZE_FOR_ALIGNMENT, BATCHRUNTOMO_CONFIG_FIDUCIALS
from .utils.binning import find_optimal_power_of_2_binning_factor
from .utils.batchruntomo import prepare_etomo_directory, run_batchruntomo
from .utils.installation import check_imod_installation


def align_tilt_series_using_fiducials(
        tilt_series: np.ndarray,
        tilt_angles: List[float],
        pixel_size: float,
        fiducial_size: float,
        nominal_rotation_angle: float,
        basename: str,
        output_directory: Path,
) -> EtomoDirectory:
    """Run fiducial based alignment in IMOD on a single tilt-series.

    Parameters
    ----------
    tilt_series: (n, y, x) array of 2D tilt-images in a tilt-series.
    tilt_angles: nominal stage tilt-angles from the microscope.
    pixel_size: nominal pixel size in Angstroms per pixel.
    fiducial_size: approximate size of fiducials in nanometers.
    nominal_rotation_angle: initial estimate for the rotation angle of the tilt
        axis. https://bio3d.colorado.edu/imod/doc/tomoguide.html#UnknownAxisAngle
    basename: basename for files in Etomo directory.
    output_directory: tilt-series directory for IMOD.
    """
    check_imod_installation()
    output_directory = Path(output_directory)
    etomo_directory = prepare_etomo_directory(
        directory=output_directory,
        tilt_series=tilt_series,
        tilt_angles=tilt_angles,
        basename=basename,

    )
    directive = generate_alignment_directive(
        tilt_series_file=etomo_directory.tilt_series_file,
        pixel_size=pixel_size,
        fiducial_size=fiducial_size,
        rotation_angle=nominal_rotation_angle
    )
    run_batchruntomo(
        directory=output_directory,
        basename=basename,
        directive=directive
    )
    if not etomo_directory.contains_alignment_results:
        raise RuntimeError(f'{basename} failed to align correctly.')
    return etomo_directory


def generate_alignment_directive(
        tilt_series_file: PathLike,
        pixel_size: float,
        fiducial_size: float,
        rotation_angle: float
) -> Dict[str, Any]:
    """Generate a fiducial-based alignment directive file for batchruntomo

    Parameters
    ----------
    tilt_series_file : file containing the tilt-series stack
    pixel_size : pixel size in the tilt-series (angstroms per pixel)
    fiducial_size : fiducial size (nanometers)
    rotation_angle : initial estimate for the rotation angle
        https://bio3d.colorado.edu/imod/doc/tomoguide.html#UnknownAxisAngle
    """
    alignment_binning_factor = find_optimal_power_of_2_binning_factor(
        src_pixel_size=pixel_size, target_pixel_size=TARGET_PIXEL_SIZE_FOR_ALIGNMENT
    )
    directive = read_adoc(BATCHRUNTOMO_CONFIG_FIDUCIALS)
    directive['setupset.copyarg.stackext'] = Path(tilt_series_file).suffix
    directive['setupset.copyarg.rotation'] = str(rotation_angle)
    directive['setupset.copyarg.pixel'] = str(pixel_size / 10)
    directive['setupset.copyarg.gold'] = str(fiducial_size)
    directive['comparam.prenewst.newstack.BinByFactor'] = str(alignment_binning_factor)
    return directive



import os

import numpy as np
from .io import read_xf


class XF:
    """Convenient retrieval of properties from IMOD xf data."""

    def __init__(self, xf: np.ndarray):
        self.xf_data = xf

    @classmethod
    def from_file(cls, filename: os.PathLike):
        return cls(read_xf(filename))

    @property
    def shifts(self):
        """Post-transformation shifts directly from xf data.

        Output is an (n, 2) numpy array of XY shifts. Shifts in an xf file are
        applied after rotations. IMOD xf files contain linear transformations. In
        the context of tilt-series alignment they contain transformations which are
        applied to 'align' a tilt-series such that images represent a fixed body rotating
        around the Y-axis.
        """
        return self.xf_data[:, -2:]

    @property
    def transformation_matrices(self):
        """2D transformation matrices directly from xf data.

        Output is an (n, 2, 2) numpy array of matrices.
        """
        return self.xf_data[:, :4].reshape((-1, 2, 2))

    @property
    def in_plane_rotations(self):
        """Extract the in plane rotation angle from IMOD xf data.

        Output is an (n, ) numpy array of angles in degrees. This assumes
        that the transformation in the xf file is a simple 2D rotation.
        """
        cos_theta = self.transformation_matrices[:, 0, 0]
        return np.rad2deg(np.arccos(cos_theta))

    @property
    def image_shifts(self):
        """Shifts to align tilt-images with projected specimen.

        Rotation center is in IMOD convention (N-1) / 2.
        """
        inverse_transformation_matrices = np.linalg.pinv(self.transformation_matrices)
        return inverse_transformation_matrices @ self.shifts.reshape((-1, 2, 1))

    @property
    def specimen_shifts(self):
        """Shifts which align projected specimen with tilt-images.

        Rotation center is in IMOD convention (N-1) / 2.
        """
        return -self.image_shifts



from pathlib import Path as _Path

BATCHRUNTOMO_CONIFG_DIR = _Path(__file__).parent / 'batchruntomo_config'
BATCHRUNTOMO_CONFIG_FIDUCIALS = BATCHRUNTOMO_CONIFG_DIR / 'fiducials.adoc'
BATCHRUNTOMO_CONFIG_PATCH_TRACKING = BATCHRUNTOMO_CONIFG_DIR / 'patch_tracking.adoc'

TARGET_PIXEL_SIZE_FOR_ALIGNMENT = 10

MINIMUM_IMOD_VERSION = '4.11.0'



from pathlib import Path


class EtomoDirectory:
    def __init__(self, basename: str, directory: Path):
        self.directory: Path = directory
        self.basename: str = basename

    @property
    def tilt_series_file(self) -> Path:
        return self.directory / f'{self.basename}.mrc'

    @property
    def rawtlt_file(self) -> Path:
        return self.directory / f'{self.basename}.rawtlt'

    @property
    def xf_file(self) -> Path:
        return self.directory / f'{self.basename}.xf'

    @property
    def tlt_file(self) -> Path:
        return self.directory / f'{self.basename}.tlt'

    @property
    def edf_file(self) -> Path:
        return self.directory / f'{self.basename}.edf'

    @property
    def is_ready_for_alignment(self) -> bool:
        return self.tilt_series_file.exists() and self.rawtlt_file.exists()

    @property
    def contains_alignment_results(self) -> bool:
        return self.xf_file.exists() and self.tlt_file.exists()

